__version__ = "1.0.0"
import kipet
from kipet.general_settings.settings import Settings
from kipet.main_modules.reaction_set import ReactionSet
from kipet.main_modules.reaction_model import ReactionModel
from kipet.input_output.kipet_io import *
kipet.settings = Settings()
